@@ -9,7 +9,6 @@
 # python -m libcst.tool print python_file.py
 
 import argparse
-import dataclasses
 import importlib
 import inspect
 import os
@@ -18,19 +17,11 @@ import shutil
 import sys
 import textwrap
 from abc import ABC, abstractmethod
-from typing import Any, Callable, Dict, List, Sequence, Tuple, Type
+from typing import Any, Callable, Dict, List, Tuple, Type
 
 import yaml
 
-from libcst import (
-    CSTNode,
-    IndentedBlock,
-    LIBCST_VERSION,
-    Module,
-    parse_module,
-    PartialParserConfig,
-)
-from libcst._nodes.deep_equals import deep_equals
+from libcst import LIBCST_VERSION, parse_module, PartialParserConfig
 from libcst._parser.parso.utils import parse_version_string
 from libcst.codemod import (
     CodemodCommand,
@@ -40,189 +31,8 @@ from libcst.codemod import (
     gather_files,
     parallel_exec_transform_with_prettyprint,
 )
-
-_DEFAULT_INDENT: str = "  "
-
-
-def _node_repr_recursive(  # noqa: C901
-    node: object,
-    *,
-    indent: str = _DEFAULT_INDENT,
-    show_defaults: bool = False,
-    show_syntax: bool = False,
-    show_whitespace: bool = False,
-) -> List[str]:
-    if isinstance(node, CSTNode):
-        # This is a CSTNode, we must pretty-print it.
-        tokens: List[str] = [node.__class__.__name__]
-        fields: Sequence["dataclasses.Field[object]"] = dataclasses.fields(node)
-
-        # Hide all fields prefixed with "_"
-        fields = [f for f in fields if f.name[0] != "_"]
-
-        # Filter whitespace nodes if needed
-        if not show_whitespace:
-
-            def _is_whitespace(field: "dataclasses.Field[object]") -> bool:
-                if "whitespace" in field.name:
-                    return True
-                if "leading_lines" in field.name:
-                    return True
-                if "lines_after_decorators" in field.name:
-                    return True
-                if isinstance(node, (IndentedBlock, Module)) and field.name in [
-                    "header",
-                    "footer",
-                ]:
-                    return True
-                if isinstance(node, IndentedBlock) and field.name == "indent":
-                    return True
-                return False
-
-            fields = [f for f in fields if not _is_whitespace(f)]
-        # Filter values which aren't changed from their defaults
-        if not show_defaults:
-
-            def _get_default(fld: "dataclasses.Field[object]") -> object:
-                if fld.default_factory is not dataclasses.MISSING:
-                    # pyre-fixme[29]: `Union[dataclasses._MISSING_TYPE,
-                    #  dataclasses._DefaultFactory[object]]` is not a function.
-                    return fld.default_factory()
-                return fld.default
-
-            fields = [
-                f
-                for f in fields
-                if not deep_equals(getattr(node, f.name), _get_default(f))
-            ]
-        # Filter out values which aren't interesting if needed
-        if not show_syntax:
-
-            def _is_syntax(field: "dataclasses.Field[object]") -> bool:
-                if isinstance(node, Module) and field.name in [
-                    "encoding",
-                    "default_indent",
-                    "default_newline",
-                    "has_trailing_newline",
-                ]:
-                    return True
-                type_str = repr(field.type)
-                if (
-                    "Sentinel" in type_str
-                    and field.name not in ["star_arg", "star", "posonly_ind"]
-                    and "whitespace" not in field.name
-                ):
-                    # This is a value that can optionally be specified, so its
-                    # definitely syntax.
-                    return True
-
-                for name in ["Semicolon", "Colon", "Comma", "Dot", "AssignEqual"]:
-                    # These are all nodes that exist for separation syntax
-                    if name in type_str:
-                        return True
-
-                return False
-
-            fields = [f for f in fields if not _is_syntax(f)]
-
-        if len(fields) == 0:
-            tokens.append("()")
-        else:
-            tokens.append("(\n")
-
-            for field in fields:
-                child_tokens: List[str] = [field.name, "="]
-                value = getattr(node, field.name)
-
-                if isinstance(value, (str, bytes)) or not isinstance(value, Sequence):
-                    # Render out the node contents
-                    child_tokens.extend(
-                        _node_repr_recursive(
-                            value,
-                            indent=indent,
-                            show_whitespace=show_whitespace,
-                            show_defaults=show_defaults,
-                            show_syntax=show_syntax,
-                        )
-                    )
-                elif isinstance(value, Sequence):
-                    # Render out a list of individual nodes
-                    if len(value) > 0:
-                        child_tokens.append("[\n")
-                        list_tokens: List[str] = []
-
-                        last_value = len(value) - 1
-                        for j, v in enumerate(value):
-                            list_tokens.extend(
-                                _node_repr_recursive(
-                                    v,
-                                    indent=indent,
-                                    show_whitespace=show_whitespace,
-                                    show_defaults=show_defaults,
-                                    show_syntax=show_syntax,
-                                )
-                            )
-                            if j != last_value:
-                                list_tokens.append(",\n")
-                            else:
-                                list_tokens.append(",")
-
-                        split_by_line = "".join(list_tokens).split("\n")
-                        child_tokens.append(
-                            "\n".join(f"{indent}{t}" for t in split_by_line)
-                        )
-
-                        child_tokens.append("\n]")
-                    else:
-                        child_tokens.append("[]")
-                else:
-                    raise Exception("Logic error!")
-
-                # Handle indentation and trailing comma.
-                split_by_line = "".join(child_tokens).split("\n")
-                tokens.append("\n".join(f"{indent}{t}" for t in split_by_line))
-                tokens.append(",\n")
-
-            tokens.append(")")
-
-        return tokens
-    else:
-        # This is a python value, just return the repr
-        return [repr(node)]
-
-
-def dump(
-    node: CSTNode,
-    *,
-    indent: str = _DEFAULT_INDENT,
-    show_defaults: bool = False,
-    show_syntax: bool = False,
-    show_whitespace: bool = False,
-) -> str:
-    """
-    Returns a string representation of the node that contains minimal differences
-    from the default contruction of the node while also hiding whitespace and
-    syntax fields.
-
-    Setting ``show_default`` to ``True`` will add fields regardless if their
-    value is different from the default value.
-
-    Setting ``show_whitespace`` will add whitespace fields and setting
-    ``show_syntax`` will add syntax fields while respecting the value of
-    ``show_default``.
-
-    When all keyword args are set to true, the output of this function is
-    indentical to the __repr__ method of the node.
-    """
-    return "".join(
-        _node_repr_recursive(
-            node,
-            indent=indent,
-            show_defaults=show_defaults,
-            show_syntax=show_syntax,
-            show_whitespace=show_whitespace,
-        )
-    )
+from libcst.display import dump, dump_graphviz
+from libcst.display.text import _DEFAULT_INDENT
 
 
 def _print_tree_impl(proc_name: str, command_args: List[str]) -> int:
@@ -252,6 +62,11 @@ def _print_tree_impl(proc_name: str, command_args: List[str]) -> int:
         action="store_true",
         help="Show values that exist only for syntax, like commas or semicolons",
     )
+    parser.add_argument(
+        "--graphviz",
+        action="store_true",
+        help="Displays the graph in .dot format, compatible with Graphviz",
+    )
     parser.add_argument(
         "--indent-string",
         default=_DEFAULT_INDENT,
@@ -286,15 +101,25 @@ def _print_tree_impl(proc_name: str, command_args: List[str]) -> int:
             else PartialParserConfig()
         ),
     )
-    print(
-        dump(
-            tree,
-            indent=args.indent_string,
-            show_defaults=args.show_defaults,
-            show_syntax=args.show_syntax,
-            show_whitespace=args.show_whitespace,
+    if not args.graphviz:
+        print(
+            dump(
+                tree,
+                indent=args.indent_string,
+                show_defaults=args.show_defaults,
+                show_syntax=args.show_syntax,
+                show_whitespace=args.show_whitespace,
+            )
+        )
+    else:
+        print(
+            dump_graphviz(
+                tree,
+                show_defaults=args.show_defaults,
+                show_syntax=args.show_syntax,
+                show_whitespace=args.show_whitespace,
+            )
         )
-    )
     return 0
 
 
